#Define Boolean Integer

Type BYTE
	Default %b * 8
End Type

Type WORD
	Default %w * 16
End Type

' Sprite2D_Op
#Define OpDefault 0
#Define OpPaint 1
#Define OpAlphaTest 2
#Define OpAlphaBlend 3
#Define OpAdd2D 4
#Define OpAdd 5
#Define OpMul 6
#Define OpMul2X 7
#Define OpMax 8
#Define OpMin 9
#Define OpBlend 10

' Sprite2D_Transform
#Define TrNone 0
#Define TrFlipX 1
#Define TrFlipY 2
#Define TrFlipXY 3
#Define TrRotCCW 4
#Define TrRotCW 5
#Define TrFlipXRotCCW 6
#Define TrFlipYRotCCW 7
#Define TrFlipXRotCW 7
#Define TrFlipYRotCW 6

' Sprite2D_FileFormat
#Define FF_RGB_BMP 0
#Define FF_RGB_PNG 1
#Define FF_RGB_JPG 2
#Define FF_ARGB_BMP 3
#Define FF_ARGB_PNG 4

' Sprite2D_ColChannel
#Define ChBlue 0
#Define ChGreen 1
#Define ChRed 2
#Define ChAlpha 3

Class CSprite2D
	
	Shared:
	
	#DllDeclare Kernel32("MulDiv")
	#DllDeclare User32("DrawText")
	#DllDeclare Ole32("CLSIDFromString")
	#DllDeclare Gdi32("SetDIBitsToDevice", "GetDIBits", "CreateFontIndirect", "GetDeviceCaps")
	#DllDeclare GdiPlus("GdiplusStartup", "GdipCreateBitmapFromFile", "GdipDisposeImage", "GdiplusShutdown", "GdipBitmapLockBits", _
	"GdipBitmapUnlockBits", "GdipGetImageWidth", "GdipGetImageHeight", "GdipCreateBitmapFromScan0", "GdipSaveImageToFile")
	
	' ImageLockMode
	#Define ImageLockModeRead &H1
	#Define ImageLockModeWrite &H2
	#Define ImageLockModeUserInputBuf &H4
	
	' Some GDI+ pixel formats
	#Define PixelFormat24bppRGB &H21808
	#Define PixelFormat32bppARGB &H26200A
	
	Type LOGFONT Align 1
		lfHeight As Integer
		lfWidth As Integer
		lfEscapement As Integer
		lfOrientation As Integer
		lfWeight As Integer
		lfItalic As BYTE
		lfUnderline As BYTE
		lfStrikeOut As BYTE
		lfCharSet As BYTE
		lfOutPrecision As BYTE
		lfClipPrecision As BYTE
		lfQuality As BYTE
		lfPitchAndFamily As BYTE
		lfFaceName As String * LF_FACESIZE
	End Type
	
	Type BITMAPINFOHEADER
		biSize As Integer
		biWidth As Integer
		biHeight As Integer
		biPlanes As WORD
		biBitCount As WORD
		biCompression As Integer
		biSizeImage As Integer
		biXPelsPerMeter As Double
		biClrUsed As Double
	End Type
	
	Type BITMAPINFO
		bmiHeader As BITMAPINFOHEADER
		bmiColors As Integer
	End Type
	
	Type BITMAP Align 2
		bmType As Integer
		bmWidth As Integer
		bmHeight As Integer
		bmWidthBytes As Integer
		bmPlanes As WORD
		bmBitsPixel As WORD
		bmBits As Integer
	End Type
	
	Type DIBSECTION
		dsBm As BITMAP
		dsBmih As BITMAPINFOHEADER
		dsBitfields[2] As Integer
		dshSection As Integer
		dsOffset As Integer
	End Type
	
	Type POINTAPI
		x As Integer
		y As Integer
	End Type
	
	Type RECT
		Left As Integer
		Top As Integer
		Right As Integer
		Bottom As Integer
	End Type
	
	Type GDIPLUSSTARTUPINPUT
		GdiplusVersion As Integer
		DebugEventCallback As Integer
		SuppressBackgroundThread As Integer
		SuppressExternalCodecs As Integer
	End Type
	
	Type BITMAPDATA
		Width As Integer
		Height As Integer
		Stride As Integer
		PixelFormat As Integer
		Scan0 As Integer
		Reserved As Integer
	End Type
	
	Type CLSID Align 1
		Data1 As Integer
		Data2 As WORD
		Data3 As WORD
		Data4[0 To 7] As BYTE
	End Type
	
	Type ENCODERPARAMETER
		GUID As CLSID
		NumberOfValues As Integer
		EncoderType As Integer
		Value As Integer
	End Type
	
	Type ENCODERPARAMETERS
		Count As Integer
		Parameter As ENCODERPARAMETER
	End Type
	'---------------------------------------------------------------
	
	Private:
	
	Dim meWidth As Integer, meHeight As Integer, meOp As Integer, meBFactor As Integer
	Dim meLeft As Integer, meRight As Integer, meTop As Integer, meBottom As Integer
	Dim cBuf[0] As Integer
	Dim bi32BitInfo As BITMAPINFO
	
	' =====================================
	Sub Initialize(%w = 1, %h = 1, %op = OpPaint)
		Init(w, h, op)
	End Sub
	
	Sub Terminate
		ReDim cBuf[0]
	End Sub
	' =====================================
	
	Sub Transform(%pSrc, %pDest, %W, %H, %Trans)
		Select Case Trans
			Case TrFlipX
				SFlipX(pSrc, pDest, W, H)
			Case TrFlipY
				SFlipY(pSrc, pDest, W, H)
			Case TrFlipXY
				SFlipXY(pSrc, pDest, W, H)
			Case TrRotCCW
				SRotCCW(pSrc, pDest, W, H)
			Case TrRotCW
				SRotCW(pSrc, pDest, W, H)
			Case TrFlipXRotCCW
				SFlipXRotCCW(pSrc, pDest, W, H)
			Case TrFlipYRotCCW
				SFlipYRotCCW(pSrc, pDest, W, H)
			Case Else
				MemMove(pDest, pSrc, W * H * 4)
		End Select
	End Sub
	
	Function SetDirLight(!X, !Y, !Z, !Brightness) As Integer
		If Brightness < 0 Then
			Brightness = 0
		ElseIf Brightness > 1 Then
			Brightness = 1
		End If
		Dim k = 127.4999 * Brightness / Sqr(X * X + Y * Y + Z * Z)
		
		Return Rgb((Z * k) + 128, (Y * k) + 128, (X * k) + 128)
	End Function
	
	Function TransformCoordAbs(%L2, %R2, %T2, %B2, %W2, _
		ByRef W, ByRef H, ByRef P1, ByRef P2) As Boolean
		Dim L, R, T, B
		
		
		If 0 >= L2 Then
			L = 0
		Else
			L = L2
		End If
		If meWidth <= R2 Then
			R = meWidth
		Else
			R = R2
		End If
		If 0 >= T2 Then
			T = 0
		Else
			T = T2
		End If
		If meHeight <= B2 Then
			B = meHeight
		Else
			B = B2
		End If
		If L >= R Then Return FALSE
		If T >= B Then Return FALSE
		
		W = R - L
		H = B - T
		
		p1 = (T * meWidth + L) * 4
		p2 = ((T - T2) * W2 + L - L2) * 4
		Return TRUE
	End Function
	
	Function TransformCoord2(%L2, %R2, %T2, %B2, %W2, _
		ByRef W, ByRef H, ByRef P1, ByRef P2) As Boolean
		Dim L, R, T, B
		
		If meLeft >= L2 Then
			L = meLeft
		Else
			L = L2
		End If
		If meRight <= R2 Then
			R = meRight
		Else
			R = R2
		End If
		If meTop >= T2 Then
			T = meTop
		Else
			T = T2
		End If
		If meBottom <= B2 Then
			B = meBottom
		Else
			B = B2
		End If
		If L >= R Then Return FALSE
		If T >= B Then Return FALSE
		
		W = R - L
		H = B - T
		
		P1 = ((T - meTop) * meWidth + L - meLeft) * 4
		P2 = ((T - T2) * W2 + L - L2) * 4
		Return TRUE
	End Function
	
	Function TransformCoord3(%L2, %R2, %T2, %B2, %L3, %R3, %T3, %B3, _
		%W2, %W3, ByRef W, ByRef H, ByRef P1, ByRef P2, ByRef P3) As Boolean
		Dim L, R, T, B
		
		If meLeft >= L2 Then
			L = meLeft
		Else
			L = L2
		End If
		If L < L3 Then L = L3
		If meRight <= R2 Then
			R = meRight
		Else
			R = R2
		End If
		If R > R3 Then R = R3
		If meTop >= T2 Then
			T = meTop
		Else
			T = T2
		End If
		If T < T3 Then T = T3
		If meBottom <= B2 Then
			B = meBottom
		Else
			B = B2
		End If
		If B > B3 Then B = B3
		If L >= R Then Return FALSE
		If T >= B Then Return FALSE
		
		W = R - L
		H = B - T
		
		P1 = ((T - meTop) * meWidth + L - meLeft) * 4
		P2 = ((T - T2) * W2 + L - L2) * 4
		P3 = ((T - T3) * W3 + L - L3) * 4
		Return TRUE
	End Function
	
	'---------------------------------------------------------------
	
	Public:
	
	Property Width() As Integer ' read-only
		Return meWidth
	End Property
	
	Property Height() As Integer ' read-only
		Return meHeight
	End Property
	
	Property DataPtr(%x, %y) As Integer ' read-only
		Return @cBuf[x + y * meWidth]
	End Property
	
	Property Op(newval = Empty) As Integer ' read/write
		If newval = Empty Then
			Return meOp
		Else
			If newval <> OpDefault Then meOp = %newval
		End If
	End Property
	
	Property BlendFactor(newval = Empty) As Integer ' read/write
		If newval = Empty Then
			Return meBFactor
		Else
			meBFactor = %newval
		End If
	End Property
	
	Method Init(%W = 0, %H = 0, %Op /*As Sprite2D_Op*/ = OpDefault)
		If Op <> OpDefault Then meOp = Op
		If Not (W And H) Then Exit Method
		If W = meWidth And H = meHeight Then Exit Method
		If W > 0 Then meWidth = W
		If H > 0 Then meHeight = H
		ClipRect(0, meWidth, 0, meHeight)
		ReDim cBuf[meWidth * meHeight]
		bi32BitInfo.bmiHeader.biBitCount = 32
		bi32BitInfo.bmiHeader.biPlanes = 1
		bi32BitInfo.bmiHeader.biSize = LenB(bi32BitInfo.bmiHeader)
		bi32BitInfo.bmiHeader.biWidth = meWidth
		bi32BitInfo.bmiHeader.biHeight = -meHeight
		bi32BitInfo.bmiHeader.biSizeImage = 4 * (meWidth * meHeight)
	End Method
	
	Method DrawLine(ByVal %lx1, ByVal %ly1, ByVal %lx2, ByVal %ly2, %c, %DotStep = 0, %IsXor = FALSE)
		Dim p1 As POINTAPI, p2 As POINTAPI
		Dim L = meLeft, R = meRight - 1, T = meTop, B = meBottom - 1, k
		
		p1.x = lx1: p1.y = ly1
		p2.x = lx2: p2.y = ly2
		
		If p1.y > p2.y Then Swap(p1, p2)
		If p1.y > B Then Exit Method
		If p2.y < T Then Exit Method
		If p1.y <> p2.y Then
			If p1.y < T Or p2.y > B Then
				k = (p2.x - p1.x) / (p2.y - p1.y)
				If p1.y < T Then
					p1.x = p2.x + (T - p2.y) * k
					p1.y = T
				End If
				If p2.y > B Then
					p2.x = p1.x + (B - p1.y) * k
					p2.y = B
				End If
			End If
		End If
		
		If p1.x > p2.x Then Swap(p1, p2)
		If p1.x > R Then Exit Method
		If p2.x < L Then Exit Method
		If p1.x <> p2.x Then
			If p1.x < L Or p2.x > R Then
				k = (p2.y - p1.y) / (p2.x - p1.x)
				If p1.x < L Then
					p1.y = p2.y + (L - p2.x) * k
					p1.x = L
				End If
				If p2.x > R Then
					p2.y = p1.y + (R - p1.x) * k
					p2.x = R
				End If
			End If
		End If
		
		SDrawDotLine(@cBuf[0], meWidth, p1, p2, (DotStep BAnd 65535) + 1, c, IsXor)
	End Method
	
	Method TileDraw(Src /*As CSprite2D*/, ByVal %dx, ByVal %dy, %W, %H, ByVal %sx = 0, ByVal %sy = 0, %Op /*As Sprite2D_Op*/ = OpDefault)
		Dim meL = meLeft, meR = meRight, meT = meTop, meB = meBottom
		Dim x, y, sw, sh
		
		Do
			If dx > meLeft Then
				If dx < meRight Then
					meLeft = dx
				Else
					Exit Do
				End If
			End If
			If dy > meTop Then
				If dy < meBottom Then
					meTop = dy
				Else
					Exit Do
				End If
			End If
			If dx + W <= meLeft Then
				Exit Do
			Else
				If dx + W < meRight Then meRight = dx + W
			End If
			If dy + H <= meTop Then
				Exit Do
			Else
				If dy + H < meBottom Then meBottom = dy + H
			End If
			sw = Src.Width(): sh = Src.Height()
			sx = sx - %Floor((sx + sw) / sw) * sw
			sy = sy - %Floor((sy + sh) / sh) * sh
			For y = dy + sy To meBottom - 1 Step sh
				For x = dx + sx To meRight - 1 Step sw
					Draw(Src, x, y, Op)
				Next
			Next
		Loop While FALSE
		
		meLeft = meL: meRight = meR
		meTop = meT: meBottom = meB
	End Method
	
	Method DrawRotate(Src /*As CSprite2D*/, ByVal %sx, ByVal %sy, ByVal %dx, ByVal %dy, ByVal !Angle, %AA = FALSE)
		Dim sw = Src.Width(), sh = Src.Height()
		Dim ddx = dx * &H10000 + &H8000, ddy = dy * &H10000 + &H8000
		Dim SinA = Sin(Angle) * &H10000, CosA = Cos(Angle) * &H10000
		Dim L = ddx - sx * CosA - sy * SinA, R = L
		Dim T = ddy - sy * CosA + sx * SinA, B = T
		Dim x = ddx - (sx - sw) * CosA - sy * SinA
		Dim yy = ddy - (sy - sh) * CosA + sx * SinA
		
		If meRight <= meLeft Then Exit Method
		If meBottom <= meTop Then Exit Method
		
		If L > x Then L = x
		If R < x Then R = x
		x = ddx - sx * CosA - (sy - sh) * SinA
		If L > x Then L = x
		If R < x Then R = x
		x = ddx - (sx - sw) * CosA - (sy - sh) * SinA
		If L > x Then L = x
		If R < x Then R = x
		
		T = ddy - sy * CosA + sx * SinA
		B = T
		Dim y = ddy - (sy - sh) * CosA + sx * SinA
		If T > y Then T = y
		If B < y Then B = y
		y = ddy - sy * CosA + (sx - sw) * SinA
		If T > y Then T = y
		If B < y Then B = y
		y = ddy - (sy - sh) * CosA + (sx - sw) * SinA
		If T > y Then T = y
		If B < y Then B = y
		
		L = L \ &H10000
		If L < meLeft Then L = meLeft
		T = T \ &H10000
		If T < meTop Then T = meTop
		R = R \ &H10000
		If R >= meRight Then
			R = meRight
		Else
			R = R + 1
		End If
		B = B \ &H10000
		If B >= meBottom Then
			B = meBottom
		Else
			B = B + 1
		End If
		If AA Then
			SDrawRotAA(Src.DataPtr(0, 0), @cBuf[0], L, T, R, B, sx, sy, dx, dy, sw, sh, meWidth, SinA, CosA)
		Else
			SDrawRot(Src.DataPtr(0, 0), @cBuf[0], L, T, R, B, sx, sy, dx, dy, sw, sh, meWidth, SinA, CosA)
		End If
	End Method
	
	Method DrawDPBM(Src /*As CSprite2D*/, ByVal %sx, ByVal %sy, ByVal %Lx, ByVal %Ly, %Lz, ByVal !Bright = !1, %PointLight = FALSE)
		Dim %W, %H, %sp, %Dp, sw = Src.Width()
		
		If Not TransformCoord2(sx, sx + sw, sy, sy + Src.Height(), sw, W, H, Dp, sp) Then Exit Method
		
		Dp = Dp + DataPtr(meLeft, meTop)
		sp = sp + Src.DataPtr(0, 0)
		
		If PointLight Then
			If meLeft < sx Then
				Lx = Lx - sx
			Else
				Lx = Lx - meLeft
			End If
			If meTop < sy Then
				Ly = Ly - sy
			Else
				Ly = Ly - meTop
			End If
			If Bright < - 2 Then
				Bright = ! - 2
			ElseIf Bright > 2 Then
				Bright = !2
			End If
			SDrawDPBMPoint(sp, Dp, W, H, sw, meWidth, Lx, Ly, Lz, Bright * &H100000)
		Else
			SDrawDPBM(sp, Dp, W, H, SetDirLight(Lx, Ly, Lz, Bright), sw, meWidth)
		End If
	End Method
	
	Method MaskDrawDPBM(Src /*As CSprite2D*/, SrcMask /*As CSprite2D*/, ByVal %sx, ByVal %sy, %Mx, %My, %Mask, _
		ByVal %Lx, ByVal %Ly, %Lz, ByVal !Bright = !1, %NotMask = FALSE, %PointLight = FALSE)
		Dim %W, %H, %sp, %Dp, %Mp, sw = Src.Width(), smw = SrcMask.Width()
		
		If Not TransformCoord3(sx, sx + sw, sy, sy + Src.Height(), Mx, Mx + smw, My, My + SrcMask.Height(), _
		sw, smw, W, H, Dp, sp, Mp) Then Exit Method
		
		Dp = Dp + DataPtr(0, 0)
		sp = sp + Src.DataPtr(0, 0)
		Mp = Mp + SrcMask.DataPtr(0, 0)
		
		If PointLight Then
			If meLeft < sx Then
				If Mx < sx Then
					Lx = Lx - sx
				Else
					Lx = Lx - Mx
				End If
			Else
				If Mx < meLeft Then
					Lx = Lx - meLeft
				Else
					Lx = Lx - Mx
				End If
			End If
			
			If meTop < sy Then
				If My < sy Then
					Ly = Ly - sy
				Else
					Ly = Ly - My
				End If
			Else
				If My < meTop Then
					Ly = Ly - meTop
				Else
					Ly = Ly - My
				End If
			End If
			
			If Bright < -2 Then
				Bright = - 2
			ElseIf Bright > 2 Then
				Bright = 2
			End If
			SMaskDrawDPBMPoint(sp, Dp, Mp, Mask, W, H, sw, meWidth, smw, Lx, Ly, Lz, Bright * &H100000, NotMask)
		Else
			SMaskDrawDPBM(sp, Dp, Mp, W, H, SetDirLight(Lx, Ly, Lz, Bright), Mask, sw, meWidth, smw, NotMask)
		End If
	End Method
	
	Method Draw(Src /*As CSprite2D*/, ByVal %sx, ByVal %sy, %Op /*As Sprite2D_Op*/ = OpDefault)
		Dim %W, %H, %sp, %Dp, sw = Src.Width()
		
		If Not TransformCoord2(sx, sx + sw, sy, sy + Src.Height(), sw, W, H, Dp, sp) Then Exit Method
		
		Dp = Dp + DataPtr(meLeft, meTop)
		sp = sp + Src.DataPtr(0, 0)
		
		If Op = OpDefault Then Op = Src.Op()
		Select Case Op
			Case OpPaint
				SPaint(sp, Dp, W, H, sw, meWidth)
			Case OpAlphaTest
				SAlphaTest(sp, Dp, W, H, sw, meWidth)
			Case OpAlphaBlend
				SAlphaBlend(sp, Dp, W, H, sw, meWidth)
			Case OpAdd2D
				SAdd2D(sp, Dp, W, H, sw, meWidth)
			Case OpAdd
				SAdd(sp, Dp, W, H, sw, meWidth)
			Case OpMul2X
				SMul2X(sp, Dp, W, H, sw, meWidth)
			Case OpMul
				SMul(sp, Dp, W, H, sw, meWidth)
			Case OpBlend
				SBlend(sp, Dp, W, H, sw, meWidth, Src.BlendFactor())
			Case OpMax
				SMax(sp, Dp, W, H, sw, meWidth)
			Case OpMin
				SMin(sp, Dp, W, H, sw, meWidth)
		End Select
	End Method
	
	Method MaskDraw(Src /*As CSprite2D*/, SrcMask /*As CSprite2D*/, ByVal %sx, ByVal %sy, _ ' sx,sy=current position of Src on destination
		%Mx, %My, %BitMask, %NotMask = FALSE, %Op /*As Sprite2D_Op*/ = OpDefault) ' mx,my=x,y where to draw SrcMask on destination
		Dim %W, %H, %sp, %Dp, %Mp, sw = Src.Width(), smw = SrcMask.Width()
		
		If Not TransformCoord3(sx, sx + sw, sy, sy + Src.Height(), Mx, Mx + smw, My, My + SrcMask.Height(), sw, smw, W, H, Dp, sp, Mp) Then Exit Method
		
		Dp = Dp + DataPtr(0, 0)
		sp = sp + Src.DataPtr(0, 0)
		Mp = Mp + SrcMask.DataPtr(0, 0)
		
		If Op = OpDefault Then Op = Src.Op()
		Select Case Op
			Case OpPaint
				SMaskPaint(sp, Dp, Mp, W, H, Mask, sw, meWidth, smw, NotMask)
			Case OpMul2X
				SMaskMul2X(sp, Dp, Mp, W, H, Mask, sw, meWidth, smw, NotMask)
			Case OpAdd2D
				SMaskAdd2D(sp, Dp, Mp, W, H, Mask, sw, meWidth, smw, NotMask)
			Case OpAdd
				SMaskAdd(sp, Dp, Mp, W, H, Mask, sw, meWidth, smw, NotMask)
			Case OpMul
				SMaskMul(sp, Dp, Mp, W, H, Mask, sw, meWidth, smw, NotMask)
			Case OpBlend
				SMaskBlend(sp, Dp, Mp, W, H, Mask, sw, meWidth, smw, NotMask, Src.BlendFactor())
			Case OpAlphaTest
				SMaskAlphaTest(sp, Dp, Mp, W, H, Mask, sw, meWidth, smw, NotMask)
			Case OpAlphaBlend
				SMaskAlphaBlend(sp, Dp, Mp, W, H, Mask, sw, meWidth, smw, NotMask)
			Case OpMax
				SMaskMax(sp, Dp, Mp, W, H, Mask, sw, meWidth, smw, NotMask)
			Case OpMin
				SMaskMin(sp, Dp, Mp, W, H, Mask, sw, meWidth, smw, NotMask)
		End Select
	End Method
	
	Method MaskIntersector(Src /*As CSprite2D*/, ByVal %sx, ByVal %sy, %Mask) As Integer
		Dim %W, %H, %sp, %Dp, sw = Src.Width()
		
		If Not TransformCoordAbs(sx, sx + sw, sy, sy + Src.Height(), sw, W, H, Dp, sp) Then Return 0
		
		Dp = Dp + DataPtr(0, 0)
		sp = sp + Src.DataPtr(0, 0)
		
		Return SMaskIntersect(sp, Dp, W, H, sw, meWidth, Mask)
	End Method
	
	Method DrawEBM(SrcBump /*As CSprite2D*/, SRCCol /*As CSprite2D*/, ByVal %sx, ByVal %sy, %DestSpace = FALSE)
		Dim %W, %H, %Dp, %sp, %Cp, sbw = SRCBump.Width()
		
		If Not TransformCoord2(sx, sx + sbw, sy, sy + SRCBump.Height(), sbw, W, H, Dp, sp) Then Exit Method
		
		Dp = Dp + DataPtr(0, 0)
		sp = sp + SRCBump.DataPtr(0, 0)
		Cp = SRCCol.DataPtr(0, 0)
		
		If DestSpace Then
			If sx < 0 Then sx = 0
			If sy < 0 Then sy = 0
			SDrawEBMEx(sp, Dp, Cp, W, H, sbw, meWidth, SRCCol.Width(), SRCCol.Height(), sx, sy, meHeight)
		Else
			SDrawEBM(sp, Dp, Cp, W, H, sbw, meWidth, SRCCol.Width(), SRCCol.Height())
		End If
	End Method
	
	Method MaskDrawEBM(SrcBump /*As CSprite2D*/, SrcMask /*As CSprite2D*/, SRCCol /*As CSprite2D*/, ByVal %sx, ByVal %sy, _
		%Mx, %My, %Mask, %NotMask = FALSE, %DestSpace = FALSE)
		Dim %W, %H, %Dp, %sp, %Mp, %Cp, sbw = SRCBump.Width(), smw = SrcMask.Width()
		
		If Not TransformCoord3(sx, sx + sbw, sy, sy + SRCBump.Height(), Mx, Mx + smw, My, My + SrcMask.Height(), _
		sbw, smw, W, H, Dp, sp, Mp) Then Exit Method
		
		Dp = Dp + DataPtr(0, 0)
		sp = sp + SRCBump.DataPtr(0, 0)
		Mp = Mp + SrcMask.DataPtr(0, 0)
		Cp = SRCCol.DataPtr(0, 0)
		
		If DestSpace Then
			If sx < 0 Then sx = 0
			If sy < 0 Then sy = 0
			SMaskDrawEBMEx(sp, Dp, Mp, Cp, Mask, W, H, sbw, meWidth, smw, SRCCol.Width(), _
			SRCCol.Height(), NotMask, sx, sy, meHeight)
		Else
			SMaskDrawEBM(sp, Dp, Mp, Cp, Mask, W, H, sbw, meWidth, smw, SRCCol.Width(), SRCCol.Height(), NotMask)
		End If
	End Method
	
	Method MulAddS2X(Src /*As CSprite2D*/, ByVal %sx, ByVal %sy, %mul, %add)
		Dim %W, %H, %sp, %Dp, sw = Src.Width()
		
		If Not TransformCoord2(sx, sx + sw, sy, sy + Src.Height(), sw, W, H, Dp, sp) Then Exit Method
		
		Dp = Dp + DataPtr(meLeft, meTop)
		sp = sp + Src.DataPtr(0, 0)
		
		SMulAdd(sp, Dp, W, H, sw, meWidth, mul, add)
	End Method
	
	Method MaskMulAddS2X(Src /*As CSprite2D*/, SrcMask /*As CSprite2D*/, ByVal %sx, ByVal %sy, _
		%Mx, %My, %mul, %add, %Mask, %NotMask = FALSE)
		Dim %W, %H, %sp, %Dp, %Mp, sw = Src.Width(), smw = SrcMask.Width()
		
		If Not TransformCoord3(sx, sx + sw, sy, sy + Src.Height(), Mx, Mx + smw, My, My + SrcMask.Height(), _
		sw, smw, W, H, Dp, sp, Mp) Then Exit Method
		
		Dp = Dp + DataPtr(0, 0)
		sp = sp + Src.DataPtr(0, 0)
		Mp = Mp + SrcMask.DataPtr(0, 0)
		
		SMaskMulAdd(sp, Dp, Mp, W, H, Mask, sw, meWidth, smw, NotMask, mul, add)
	End Method
	
	Method MoveByte(Src /*As CSprite2D*/, ByVal %sx, ByVal %sy, %chSrc /*As Sprite2D_ColChannel*/, %chDest /*As Sprite2D_ColChannel*/)
		Dim %W, %H, %sp, %Dp, sw = Src.Width()
		
		If Not TransformCoord2(sx, sx + sw, sy, sy + Src.Height(), sw, W, H, Dp, sp) Then Exit Method
		
		Dp = Dp + DataPtr(meLeft, meTop)
		sp = sp + Src.DataPtr(0, 0)
		
		SMoveByte(sp, Dp, W, H, sw, meWidth, chSrc, chDest)
	End Method
	
	Method MoveBit(Src /*As CSprite2D*/, ByVal %sx, ByVal %sy, %mbSrc, %mbDest)
		Dim %W, %H, %sp, %Dp, sw = Src.Width()
		
		If Not TransformCoord2(sx, sx + sw, sy, sy + Src.Height(), sw, W, H, Dp, sp) Then Exit Method
		
		Dp = Dp + DataPtr(meLeft, meTop)
		sp = sp + Src.DataPtr(0, 0)
		
		SMoveBit(sp, Dp, W, H, sw, meWidth, mbSrc, mbDest)
	End Method
	
	Method MaskMoveByte(Src /*As CSprite2D*/, SrcMask /*As CSprite2D*/, ByVal %sx, ByVal %sy, _
		%Mx, %My, %chSrc /*As Sprite2D_ColChannel*/, %chDest /*As Sprite2D_ColChannel*/, %Mask, %NotMask = FALSE)
		Dim %W, %H, %sp, %Dp, %Mp, sw = Src.Width(), sh = Src.Height(), smw = SrcMask.Width()
		
		If Not TransformCoord3(sx, sx + sw, sy, sy + sh, Mx, Mx + smw, My, My + SrcMask.Height(), sw, smw, W, H, _
		Dp, sp, Mp) Then Exit Method
		
		Dp = Dp + DataPtr(0, 0)
		sp = sp + Src.DataPtr(0, 0)
		Mp = Mp + SrcMask.DataPtr(0, 0)
		
		SMaskMoveByte(sp, Dp, Mp, W, H, Mask, sw, meWidth, smw, NotMask, chSrc, chDest)
	End Method
	
	Sub MaskMoveBit(Src /*As CSprite2D*/, SrcMask /*As CSprite2D*/, ByVal %sx, ByVal %sy, _
		%Mx, %My, %mbSrc, %mbDest, %Mask, %NotMask = FALSE)
		Dim %W, %H, %sp, %Dp, %Mp, sw = Src.Width(), sh = Src.Height(), smw = SrcMask.Width()
		
		If Not TransformCoord3(sx, sx + sw, sy, sy + sh, Mx, Mx + smw, My, My + SrcMask.Height(), sw, smw, W, H, _
		Dp, sp, Mp) Then Exit Method
		
		Dp = Dp + DataPtr(0, 0)
		sp = sp + Src.DataPtr(0, 0)
		Mp = Mp + SrcMask.DataPtr(0, 0)
		
		SMaskMoveBit(sp, Dp, Mp, W, H, Mask, sw, meWidth, smw, NotMask, mbSrc, mbDest)
	End Method
	
	Method ClearBuffer(%c)
		If meRight <= meLeft Then Exit Method
		If meBottom <= meTop Then Exit Method
		SClearC(@cBuf[meLeft + meTop * meWidth], meRight - meLeft, meBottom - meTop, meWidth, c)
	End Method
	
	Method MaskClearBuffer(SrcMask /*As CSprite2D*/, %Mx, %My, %c, %Mask, %NotMask = FALSE)
		Dim %W, %H, %Mp, %Dp, sw = SrcMask.Width()
		
		If Not TransformCoord2(Mx, Mx + sw, My, My + SrcMask.Height(), sw, W, H, Dp, Mp) Then Exit Method
		
		Dp = Dp + DataPtr(meLeft, meTop)
		Mp = Mp + SrcMask.DataPtr(0, 0)
		
		SMaskClearC(Dp, Mp, W, H, Mask, meWidth, sw, c, NotMask)
	End Method
	
	Method LoadFromTGA($fName, %ColorKey = -1)
		Dim w, h, X, Y, t, %d
		
		If FileExist(fName) Then
			FileGet(FileOpen(fname, BINARY), FileLen(fName))
			FileClose(FileOpen)
		Else
			Exit Method
		End If
		
		Select Case Peek(@FileGet + 16, 1)
			Case &H18: t = 3: meOp = OpPaint
			Case &H20: t = 4: meOp = OpAlphaBlend
			Case Else: FreePtr(@FileGet): Exit Method
		End Select
		
		If ColorKey >= 0 Then meOp = OpAlphaTest
		w = Peek(@FileGet + 12, 2)
		h = Peek(@FileGet + 14, 2)
		Init(w, h, meOp)
		Dim m As String * (w * t)
		d = &HFF000000
		For Y = 0 To h - 1
			MemMove(@m, @FileGet + 18 + w * t * Y, LenB(m))
			For X = 0 To w - 1
				Poke(@d, Peek(@m + X * t, t))
				If meOp = OpAlphaTest Then
					If (d BAnd &HFFFFFF) = ColorKey Then
						cBuf[X + (h - Y - 1) * w] = ColorKey
					Else
						cBuf[X + (h - Y - 1) * w] = d BOr &HFF000000
					End If
				Else
					cBuf[X + (h - Y - 1) * w] = d
				End If
			Next
		Next
		FreePtr(@FileGet)
	End Method

	Method LoadFromFile($fName, %Trans /*As Sprite2D_Transform*/ = TrNone, %W = 0, %H = 0, %cKey = -1)
		Dim tSI As GDIPLUSSTARTUPINPUT, lData As BITMAPDATA, R As RECT
		Dim %Res, %GDIP, %Bmp, %x, %y
		
		If (Trans <> TrNone) OrElse (W > 0) OrElse (H > 0) Then
			Dim sp As New CSprite2D
			sp.LoadFromFile(fName)
			LoadFromSprite(sp, Trans, W, H, cKey)
			sp.Dispose()
		Else
			tSI.GdiplusVersion = 1
			Res = GdiplusStartup(@GDIP, @tSI)
			If Res = 0 Then
				Res = GdipCreateBitmapFromFile(AnsiToWide(fName), @Bmp): FreePtr(AnsiToWide)
				If Res = 0 Then
					GdipGetImageWidth(Bmp, @x)
					GdipGetImageHeight(Bmp, @y)
					R.Right = x: R.Bottom = y
					Res = GdipBitmapLockBits(Bmp, @R, ImageLockModeRead, PixelFormat32bppARGB, @lData)
					If Res = 0 Then
						Init(x, y, OpPaint)
						MemMove(@cBuf[0], lData.Scan0, x * y * 4)
						GdipBitmapUnlockBits(Bmp, @lData)
						If cKey >= 0 Then
							AddColorKey(cKey)
							Op(OpAlphaTest)
						End If
					End If
					GdipDisposeImage(Bmp)
				End If
				GdiplusShutdown(GDIP)
			End If
		End If
	End Method
	
	Method LoadFromSprite(Src /*As CSprite2D*/, Trans /*As Sprite2D_Transform*/ = TrNone, ByVal %W = 0, ByVal %H = 0, %cKey = -1)
		Dim fResize = FALSE
		
		If W <= 0 Then
			W = Src.Width()
		Else
			If W <> Src.Width() Then fResize = TRUE
		End If
		If H <= 0 Then
			H = Src.Height()
		Else
			If H <> Src.Height() Then fResize = TRUE
		End If
		If fResize Then
			Init(W, H, Src.Op())
			If Trans = TrNone Then
				SResize(Src.DataPtr(0, 0), DataPtr(0, 0), Src.Width(), Src.Height(), meWidth, meHeight)
			ElseIf W * H > Src.Width() * Src.Height() Then
				W = Src.Width: H = Src.Height
				Dim Ar As String * (W * H - 1)
				Transform(Src.DataPtr(0, 0), @Ar, W, H, Trans)
				If Trans > TrFlipXY Then Swap(W, H)
				SResize(@Ar, @cBuf[0], W, H, meWidth, meHeight)
			Else
				Dim Ar As String * (W * H - 1)
				If Trans > TrFlipXY Then Swap(W, H)
				SResize(Src.DataPtr(0, 0), @Ar, Src.Width, Src.Height, W, H)
				Transform(@Ar, @cBuf[0], W, H, Trans)
			End If
		Else
			If Trans > TrFlipXY Then Swap(W, H)
			Init(W, H, Src.Op())
			Transform(Src.DataPtr(0, 0), @cBuf[0], Src.Width(), Src.Height(), Trans)
		End If
		If cKey >= 0 Then
			AddColorKey(cKey)
			Op(OpAlphaTest)
		End If
	End Method
	
	Method LoadText($strtodisp, %fcolor = &HFFFFFF, %bcolor = 0, %ColorKey = -1, %ptsize = 0, %bold = FALSE, %italic = FALSE, %underline = FALSE, $face = "")
		Dim rc As RECT, font As LOGFONT
		Dim X, Y, w, h, height, cur
		Dim hBmp = NULL, hFont = NULL
		
		font.lfHeight = -MulDiv(ptsize, GetDeviceCaps(CreateCompatibleDC(GetDC(ME)), 90), 72) ' LOGPIXELSY
		font.lfWeight = IIf(bold, FW_EXTRABOLD, FW_NORMAL)
		font.lfItalic = italic
		font.lfUnderline = underline
		font.lfFaceName = face
		hFont = CreateFontIndirect(@font)
		If Not hFont Then GoTo TxtError
		
		hFont = SelectObject(CreateCompatibleDC, hFont)
		If Not DrawText(CreateCompatibleDC, strtodisp, -1, @rc, &H405) Then GoTo TxtError ' DT_CENTER+DT_VCENTER+DT_CALCRECT
		
		w = rc.Right: h = rc.Bottom
		
		hBmp = SelectObject(CreateCompatibleDC, CreateCompatibleBitmap(GetDC, w, h))
		FillStyle(FILL_SOLID): Fill(CreateCompatibleDC, 1, 1, bcolor)
		ForeColor(CreateCompatibleDC, fcolor): FontTransparent(CreateCompatibleDC, TRUE)
		If Not DrawText(CreateCompatibleDC, strtodisp, -1, @rc, 5) Then GoTo TxtError ' DT_CENTER+DT_VCENTER
		
		If ColorKey >= 0 Then
			Init(w, h, OpAlphaTest)
		Else
			Init(w, h, OpPaint)
		End If
		
		If Not GetDIBits(CreateCompatibleDC, CreateCompatibleBitmap, 0, meHeight, @cBuf[0], @bi32BitInfo, 0) Then GoTo TxtError
		
		If ColorKey >= 0 Then
			For Y = 0 To h - 1
				height = Y * meWidth
				For X = 0 To w - 1
					cur = X + height
					If (cBuf[cur] BAnd &HFFFFFF) = ColorKey Then
						cBuf[cur] = ColorKey
					Else
						cBuf[cur] = cBuf[cur] BOr &HFF000000
					End If
				Next
			Next
		Else
			For Y = 0 To h - 1
				height = Y * meWidth
				For X = 0 To w - 1
					cur = X + height
					cBuf[cur] = cBuf[cur] BOr &HFF000000
				Next
			Next
		End If
		
		:Error
		ReleaseDC(ME, GetDC)
		DeleteObject(SelectObject(CreateCompatibleDC, hFont))
		DeleteObject(SelectObject(CreateCompatibleDC, hBmp))
		DeleteDC(CreateCompatibleDC)
	End Method
	
	Method SaveToFile($fName, %fFormat, %JPGQuality = 80)
		Dim tSI As GDIPLUSSTARTUPINPUT
		Dim tEncoder As CLSID
		Dim tParams As ENCODERPARAMETERS
		Dim %Res, %GDIP, %Bmp, %Stride, $sType
		
		tSI.GdiplusVersion = 1
		Res = GdiplusStartup(@GDIP, @tSI)
		If Res = 0 Then
			If (fFormat = FF_ARGB_BMP) OrElse (fFormat = FF_ARGB_PNG) OrElse (fFormat = FF_RGB_JPG) Then
				Res = GdipCreateBitmapFromScan0(Width, Height, Width * 4, PixelFormat32bppARGB, @cBuf[0], @Bmp)
			Else
				Stride = ((Width * 3 - 1) BOr 3) + 1
				Dim ArRGB As String * (Stride * Height)
				Bpp32to24(@cBuf[0], @ArRGB, Width, Height, Stride)
				Res = GdipCreateBitmapFromScan0(Width, Height, Stride, PixelFormat24bppRGB, @ArRGB, @Bmp)
			End If
			If Res = 0 Then
				Select Case fFormat
					Case FF_ARGB_BMP, FF_RGB_BMP
						sType = "{557CF400-1A04-11D3-9A73-0000F81EF32E}"
					Case FF_ARGB_PNG, FF_RGB_PNG
						sType = "{557CF406-1A04-11D3-9A73-0000F81EF32E}"
					Case FF_RGB_JPG
						sType = "{557CF401-1A04-11D3-9A73-0000F81EF32E}"
						tParams.Count = 1
						tParams.Parameter.NumberOfValues = 1
						tParams.Parameter.EncoderType = 4
						tParams.Parameter.Value = @JPGQuality
						CLSIDFromString(AnsiToWide("{1D5BE4B5-FA4A-452D-9CDD-5DB35105E7EB}"), _
						@tParams.Parameter.GUID): FreePtr(AnsiToWide)
				End Select
				CLSIDFromString(AnsiToWide(sType), @tEncoder): FreePtr(AnsiToWide)
				If tParams.Count = 0 Then
					Res = GdipSaveImageToFile(Bmp, AnsiToWide(fName), @tEncoder, NULL): FreePtr(AnsiToWide)
				Else
					Res = GdipSaveImageToFile(Bmp, AnsiToWide(fName), @tEncoder, @tParams): FreePtr(AnsiToWide)
				End If
				GdipDisposeImage(Bmp)
			End If
			GdiplusShutdown(GDIP)
		End If
	End Method
	
	Method ClipRect(%L = -1, %R = -1, %Top = -1, %Bottom = -1)
		If L >= 0 And L < meWidth Then
			meLeft = L
		Else
			meLeft = 0
		End If
		If R > L And R <= meWidth Then
			meRight = R
		Else
			meRight = meWidth
		End If
		If Top >= 0 And Top < meHeight Then
			meTop = Top
		Else
			meTop = 0
		End If
		If Bottom > Top And Bottom <= meHeight Then
			meBottom = Bottom
		Else
			meBottom = meHeight
		End If
	End Method
	
	Method AddColorKey(%cKey)
		SAddColorKey(@cBuf[0], meWidth, meHeight, cKey)
	End Method
	
	Method PaintToDevice(%hDC)
		SetDIBitsToDevice(hDC, 0, 0, meWidth, meHeight, 0, 0, 0, meHeight, @cBuf[0], @bi32BitInfo, 0)
	End Sub
	
	Method ClearRect(ByVal %Left, ByVal %Right, ByVal %Top, ByVal %Bottom, %c)
		If Left < meLeft Then Left = meLeft
		If Right > meRight Then Right = meRight
		If Right <= Left Then Exit Method
		If Top < meTop Then Top = meTop
		If Bottom > meBottom Then Bottom = meBottom
		If Bottom <= Top Then Exit Method
		SClearC(@cBuf[Left + Top * meWidth], Right - Left, Bottom - Top, meWidth, c)
	End Method
	
	Method SetPixel(%X, %Y, %c)
		If X < 0 Then Exit Method
		If X >= meWidth Then Exit Method
		If Y < 0 Then Exit Method
		If Y >= meHeight Then Exit Method
		cBuf[X + Y * meWidth] = c
	End Method
	
	Method GetPixel(%X, %Y) As Integer
		If X < 0 Then Exit Method
		If X >= meWidth Then Exit Method
		If Y < 0 Then Exit Method
		If Y >= meHeight Then Exit Method
		Return cBuf[X + Y * meWidth]
	End Method
	
End Class